// I2Cdev and MPU6050 must be installed as libraries, or else the .cpp/.h files
// for both classes must be in the include path of your project
#include "I2Cdev.h"

#include "MPU6050_6Axis_MotionApps20.h"
//#include "MPU6050.h" // not necessary if using MotionApps include file

// Arduino Wire library is required if I2Cdev I2CDEV_ARDUINO_WIRE implementation
// is used in I2Cdev.h
#if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
    #include "Wire.h"
#endif

// class default I2C address is 0x68
// specific I2C addresses may be passed as a parameter here
// AD0 low = 0x68 (default for SparkFun breakout and InvenSense evaluation board)
// AD0 high = 0x69
MPU6050 mpu;
//MPU6050 mpu(0x69); // <-- use for AD0 high

#define LED_PIN 13 // (Arduino is 13, Teensy is 11, Teensy++ is 6)
bool blinkState = false;

int recal; // Recalibration flag
int loopCtr; // Loop counter flag

// MPU control/status vars
bool dmpReady = false;  // set true if DMP init was successful
uint8_t mpuIntStatus;   // holds actual interrupt status byte from MPU
uint8_t devStatus;      // return status after each device operation (0 = success, !0 = error)
uint16_t packetSize;    // expected DMP packet size (default is 42 bytes)
uint16_t fifoCount;     // count of all bytes currently in FIFO
uint8_t fifoBuffer[64]; // FIFO storage buffer

// orientation/motion vars
Quaternion q;           // [w, x, y, z]         quaternion container
VectorInt16 aa;         // [x, y, z]            accel sensor measurements
VectorInt16 aaReal;     // [x, y, z]            gravity-free accel sensor measurements
VectorInt16 aaWorld;    // [x, y, z]            world-frame accel sensor measurements
VectorFloat gravity;    // [x, y, z]            gravity vector
float euler[3];         // [psi, theta, phi]    Euler angle container
float ypr[3];           // [yaw, pitch, roll]   yaw/pitch/roll container and gravity vector
float zCalibrate;
float  calibrationQuality = 0.1; // DEGREES
float B_U; // Bottom - Up angle required by client
float L_R; // Left - Right angle required by client
float LR_Meas; // Buffer for L-R values read
float temp; // Temporary buffer
char buf[10];

void calibrateAngles(); // Calibration function prototype

// ================================================================
// ===                      INITIAL SETUP                       ===
// ================================================================

void setup() {
  // join I2C bus (I2Cdev library doesn't do this automatically)
  #if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
      Wire.begin();
      Wire.setClock(400000); // 400kHz I2C clock. Comment this line if having compilation difficulties
  #elif I2CDEV_IMPLEMENTATION == I2CDEV_BUILTIN_FASTWIRE
      Fastwire::setup(400, true);
  #endif

  // initialize serial communication
  // (115200 chosen because it is required for Teapot Demo output, but it's
  // really up to you depending on your project)
  Serial.begin(115200);
  while (!Serial); // wait for Leonardo enumeration, others continue immediately
  // NOTE: 8MHz or slower host processors, like the Teensy @ 3.3V or Arduino
  // Pro Mini running at 3.3V, cannot handle this baud rate reliably due to
  // the baud timing being too misaligned with processor ticks. You must use
  // 38400 or slower in these cases, or use some kind of external separate
  // crystal solution for the UART timer.

  // initialize device
  mpu.initialize();
    
  // load and configure the DMP
  devStatus = mpu.dmpInitialize();

  // Calibration Time: generate offsets and calibrate our MPU6050
  // calibrateAngles();
  //mpu.CalibrateAccel(6);
  mpu.CalibrateGyro(6);
  
  // turn on the DMP, now that it's ready
  mpu.setDMPEnabled(true);

  mpuIntStatus = mpu.getIntStatus();

  // set our DMP Ready flag so the main loop() function knows it's okay to use it
  dmpReady = true;

  // get expected DMP packet size for later comparison
  packetSize = mpu.dmpGetFIFOPacketSize();

  // configure LED for output
  pinMode(LED_PIN, OUTPUT);
}


// ================================================================
// ===                    MAIN PROGRAM LOOP                     ===
// ================================================================

void loop() {
  
    // if programming failed, don't try to do anything
    if (!dmpReady) return;
    // read a packet from FIFO
    if (mpu.dmpGetCurrentFIFOPacket(fifoBuffer)) { // Get the Latest packet
      // Look for calibration request
      if (Serial.available()){
        recal = Serial.read();
        if (recal){
          calibrateAngles();
        }
      }

      mpu.dmpGetQuaternion(&q, fifoBuffer);
      mpu.dmpGetGravity(&gravity, &q);
      mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);
      
      if (loopCtr >= 50) {
        getBottomUpAngle(-ypr[2], -ypr[1]);
        Serial.println("|");
        delay(50);
        Serial.print("VE: ");
        Serial.print(B_U * 180/M_PI,1);
        delay(50);
        Serial.print(" / L-R: ");
        LR_Meas = ypr[0]*180/M_PI;
        temp = LR_Meas - zCalibrate;
        if(zCalibrate >= 0)
          if(LR_Meas >= 0 && LR_Meas >= zCalibrate)
            temp = temp;
          else if(LR_Meas >= 0 && LR_Meas < zCalibrate)
            temp = temp + 360.0;
          else // LR_Meas < 0
            temp = temp + 360.0;
        else
          if(LR_Meas <= 0 && LR_Meas >= zCalibrate)
            temp = temp;
          else if(LR_Meas <= 0 && LR_Meas < zCalibrate)
            temp = 360 - temp;
          else
            temp = temp;
        Serial.print(360-temp, 1);
        loopCtr = 0;
      }
    }

    loopCtr++;
    delay(2);
}

void calibrateAngles(){
  // Calibrate all axes
  //mpu.CalibrateAccel(6);
  mpu.CalibrateGyro(6);
  zCalibrate = 0;
  bool calibrateFlag = true;
  while (calibrateFlag){
    calibrateFlag = false;
    zCalibrate = ypr[0];
    // Scan values for 2 seconds
    for(int i = 0; i < 20; i++){
      mpu.dmpGetCurrentFIFOPacket(fifoBuffer);
      mpu.dmpGetQuaternion(&q, fifoBuffer);
      mpu.dmpGetGravity(&gravity, &q);
      mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);
      //Serial.print(ypr[0]*180.0/M_PI,1);
      //Serial.print(" VS ");
      //Serial.print(zCalibrate*180.0/M_PI,1);
      //Serial.print("\tdifference: ");
      //Serial.println(abs(zCalibrate - ypr[0])*180.0/M_PI,2);
      if (abs(zCalibrate - ypr[0])*180.0/M_PI > calibrationQuality){
        calibrateFlag = true;
        //break;
      }
      delay(100); // Avoid repeat measurements
    }
  }

  // Even out readings over 500 runs
  for(int i = 0; i<500; i++){
    mpu.dmpGetQuaternion(&q, fifoBuffer);
    mpu.dmpGetGravity(&gravity, &q);
    mpu.dmpGetYawPitchRoll(ypr, &q, &gravity);
    zCalibrate += ypr[0];
    delay(3);
  }
  zCalibrate *= 180.0/M_PI/500.0;
  
  recal = 0;
}

float getBottomUpAngle(float alpha, float beta){
  float x = cos(alpha) * cos(beta);
  float z = sin(alpha) * cos(beta);
  float y = sin(beta);
  B_U = atan(y/sqrt(x*x+z*z));
}
